%% Deterministic and stochastic Maxwell 2d
%% From Component splitting for semi-discrete Maxwell equations, Verwer
%% Stagged uniform grid
%% PEC+PMC 
%% Parameters similar to Chen, Hong, Zhang 2016
%% Stochastic exponential method, BEM, EM
%% trace formula for energy and preservation of divergence

clear all
rng(113)

fprintf('traceMaxwell2dVerwer.m\n');

% parameter in front of the noise
% yesno=0 -> no noise
yesno=.5; 

% number of samples
Ms=25000

% time discretisation
T=5;
N=500;
dt=T/N;
tt=[0+dt:dt:T];

% space discretisation wlog Mx=My and dx=dy
Il=0;Ir=1; % x interval
M=2^4; % 2^4 takes a lot of time  
dx=(Ir-Il)/(M); 
xx=dx*[1:M-1]'; 
xx12=[Il+dx/2:dx:Ir-dx/2];
Ily=0;Iry=1; % y interval
My=M; 
dy=(Iry-Ily)/(My);
yy=dy*[1:My-1]'; 
yy12=[Il+dy/2:dy:Ir-dy/2];

% FD matrices d/dx or d/dy assuming dx=dy !! 
AEy=sparse(M*(M-1),(M-1)^2);
AEx=AEy;
for j=1:(M-1)^2
  AEy(j,j)=1;
end
for j=1:(M-1)^2
  AEy(j+(M-1),j)=-1;
end
AA=spdiags([-1*ones(M-1,1) ones(M-1,1)],[-1:0],M,M-1);
Ix=speye(M-1); 
AEx=kron(Ix,AA);
AHy=-AEy.';
AHx=-AEx.';
zorro1=spalloc((M-1)^2,(M-1)^2,0);
zorro2=spalloc(M*(M-1),M*(M-1),0);
% approximation of matrix [0 -d/dy d/dx; -d/dy 0 0; d_x 0 0]
B=[ zorro1  -AHy +AHx; -AEy zorro2 zorro2; AEx zorro2 zorro2 ]; 
% same dx and dy
B=B/dx; 

expA=expm(dt*B); 
Bbem=speye((M-1)^2+2*M*(M-1))-dt*B; % matrix for BEM

% initial conditions
E3=zeros(M-1,M-1); 
H1=zeros(M-1,M); 
H2=zeros(M,M-1);
% E3 must be zero on the boundary
for j=1:M-1
  for k=1:M-1
      %% pulse
      E3(j,k)=.1*exp(-50*((xx(j)-.5).^2+(yy(k)-.5).^2));
      %% chz16.pdf
      %E3(j,k)=sin(2*pi*xx(j))*sin(2*pi*yy(k))*.3;
      %% rdm
      %E3(j,k)=randn*sin(2*pi*xx(j))*sin(2*pi*yy(k));
   end
end   
%E3, stop
%surf(E3), stop


% H1 is constant in x
%  for j=1:M-1
%    for k=1:M
%        H1(j,k)=1*xx(j)+0*yy12(k);
%    end
%  end   
for j=1:M-1
  H1(j,:)=ones(M,1)*randn;
end  
%H1

% H2 is constant in y
%  for j=1:M
%    for k=1:M-1
%        H2(j,k)=0*xx12(j)+1*yy(k);
%     end
%  end  
for k=1:M-1
  H2(:,k)=ones(M,1)*randn;
end  
%H2

Uinit=[E3(:); H1(:); H2(:)];

% initial energy and divergences
energy0=dx*dy*sum(Uinit.^2); 
% divE0=0 because E=(0,0,E3(x,y))
% divH0
% reshape H1 and H2 first
H1t=reshape(H1.',1,[])'; H2t=reshape(H2.',1,[])';
divH0=AHx*H1t+AHy*H2t; norm(divH0), 

ePlot=zeros(N,1);
ePlotem=ePlot;ePlotbem=ePlot;
dhPlot=zeros(N, (M-1)^2);
dhPlotbem=dhPlot; dhPlotem=dhPlot;

% eigenvalues for the noise from chz16.pdf more or less
E3_tempo=zeros(M-1);
for j=1:M-1
  for k=1:M-1
      E3_tempo(j,k)=3/(j^3+k^3);
   end
end   
for j=1:M-1
  for k=0:M-1
      H1_tempo(j,k+1)=3/(j^3+(k+1/2)^3);
   end
end   
for j=0:M-1
  for k=1:M-1
      H2_tempo(j+1,k)=3/((j+1/2)^3+k^3);
   end
end   
E3tempo=reshape(E3_tempo,1,(M-1)^2);
H1tempo=reshape(H1_tempo,1,M*(M-1));
H2tempo=reshape(H2_tempo,1,M*(M-1));
E_ex=[1.*E3tempo 1*H1tempo 1*H2tempo]'; 

for m=1:Ms
 % init
  U1=Uinit; 
  U1em=Uinit;
  U1bem=Uinit;

  % noise 
  Wtt=randn(N,(M-1)^2+2*M*(M-1));
  W=zeros(N,(M-1)^2+2*M*(M-1));
  for j=1:(M-1)^2+2*M*(M-1)
    W(:,j)=E_ex(j)*Wtt(:,j);
  end
  V=zeros(N,(M-1)^2+2*M*(M-1));
  for i=1:N
    V(i,:)=sum(W(1*(i-1)+1:1*i,:),1);
  end  
  Wn=yesno*sqrt(dt)*V;
  
  % time discretisation
  for j=1:N
  
    %% SEXP integrator
    U1=expA*( U1+Wn(j,:)' );
    
    % energy and divergence
    E3tsexp=U1(1:(M-1)^2);
    H1tsexp=U1((M-1)^2+1:(M-1)^2+M*(M-1));
    H2tsexp=U1((M-1)^2+M*(M-1)+1:end); 
        
    energyy=dx*dy*sum(U1.^2); 
    ePlot(j)=ePlot(j)+energyy;

    % need to reshape H1 and H2
    tempo1=reshape(H1tsexp.',[M-1,M]).'; 
    tempo2=reshape(H2tsexp.',[M,M-1]).';
    Tempo1=tempo1(:); Tempo2=tempo2(:);
    divH=AHx*Tempo1+AHy*Tempo2;
    dhPlot(j,:)=dhPlot(j,:)+divH'; 
    %dhPlot(j)=dhPlot(j)+sqrt(dx)*norm(divH-divH0);
    
    
       %E3plot=reshape(E3tsexp,M-1,M-1);
       %surf(E3plot),drawnow,
       %E3plot=reshape(H1tsexp,M-1,M);
       %surf(E3plot),drawnow,
       %E3plot=reshape(H2tsexp,M,M-1);
       %surf(E3plot),drawnow,
       
     %% EM integrator
     U1em=U1em+dt*B*U1em+Wn(j,:)';
     
     % energy and divergences
     E3tem=U1em(1:(M-1)^2);
     H1tem=U1em((M-1)^2+1:(M-1)^2+M*(M-1));
     H2tem=U1em((M-1)^2+M*(M-1)+1:end); 
    
     energyem=dx*dy*sum(U1em.^2);     
     ePlotem(j)=ePlotem(j)+energyem;
     
     % need to reshape H1 and H2
     tempo1=reshape(H1tem.',[M-1,M]).';
     tempo2=reshape(H2tem.',[M,M-1]).';
     Tempo1=tempo1(:); Tempo2=tempo2(:);
     divHem=AHx*Tempo1+AHy*Tempo2; 
     dhPlotem(j,:)=dhPlotem(j,:)+divHem';
     %dhPlotem(j)=dhPlotem(j)+sqrt(dx)*norm(divHem-divH0);
     
       %E3plot=reshape(E3tem,M-1,M-1);
       %surf(E3plot),drawnow,
       %E3plot=reshape(H1tem,M-1,M);
       %surf(E3plot),drawnow,
       %E3plot=reshape(H2tem,M,M-1);
       %surf(E3plot),drawnow
       
     %% BEM integrator
     U1bem=Bbem\( U1bem+Wn(j,:)' );
     
     % energy and divergences
     E3tbem=U1bem(1:(M-1)^2);
     H1tbem=U1bem((M-1)^2+1:(M-1)^2+M*(M-1));
     H2tbem=U1bem((M-1)^2+M*(M-1)+1:end); 

     energybem=dx*dy*sum(U1bem.^2); 
     ePlotbem(j)=ePlotbem(j)+energybem;
     
     % need to reshape H1 and H2
     tempo1=reshape(H1tbem.',[M-1,M]).';
     tempo2=reshape(H2tbem.',[M,M-1]).';
     Tempo1=tempo1(:); Tempo2=tempo2(:);
     divHbem=AHx*Tempo1+AHy*Tempo2; 
     dhPlotbem(j,:)=dhPlotbem(j,:)+divHbem';
     %dhPlotbem(j)=dhPlotbem(j)+sqrt(dx)*norm(divHbem-divH0);
     
     
       %E3plot=reshape(E3tbem,M-1,M-1);
       %surf(E3plot),drawnow
       %E3plot=reshape(H1tbem,M-1,M-1);
       %surf(E3plot),drawnow
       %E3plot=reshape(H2tbem,M-1,M-1);
       %surf(E3plot),drawnow
  end % N
  %m
end % Ms

set(0,'DefaultTextFontSize',12)
set(0,'DefaultAxesFontSize',12)
set(0,'DefaultLineLineWidth',2)
set(0,'DefaultLineMarkerSize',10)

%% plot trace formula
everystp=45;
ttplot=tt(1:everystp:end);
traceQ=dx*dy*( sum(E_ex.^2) ); 
figure(),plot(ttplot,ePlot(1:everystp:end)/Ms,'ks-',ttplot,ePlotbem(1:everystp:end)/Ms,'k+-',tt,energy0+yesno^2*tt*traceQ,'k--')
xlabel('Time','FontSize',14);
ylabel('Energy','Rotation',0,'HorizontalAlignment','right','FontSize',14);
legend('SEXP','SEM','Exact','Location','NorthWest');
%  str=sprintf('$M_s=$ %d, $\\Delta t=$ %.2f, $M=$ %d',Ms,dt,M);
%  title(str,'Interpreter','latex','FontSize',14)
print -depsc2 -r0 traceMaxwell2dV-R.eps

%% plot trace formula zoom for EM
zzoom=450;
traceQ=dx*dy*( sum(E_ex.^2) ); 
everystp=10;
ttplot=tt(1:everystp:end);
figure(),plot(tt(1:everystp:end-zzoom),ePlot(1:everystp:end-zzoom)/Ms,'ks-',tt(1:everystp:end-zzoom),ePlotbem(1:everystp:end-zzoom)/Ms,'k+-', ... 
            tt(1:everystp:end-zzoom),ePlotem(1:everystp:end-zzoom)/Ms,'kd-',tt(1:everystp:end-zzoom),energy0+yesno^2*tt(1:everystp:end-zzoom)*traceQ,'k--')
xlabel('Time','FontSize',14);
ylabel('Energy','Rotation',0,'HorizontalAlignment','right','FontSize',14);
legend('SEXP','SEM','EM','Exact','Location','NorthWest');
%  str=sprintf('$M_s=$ %d, $\\Delta t=$ %.2f, $M=$ %d',Ms,dt,M);
%  title(str,'Interpreter','latex','FontSize',14)
print -depsc2 -r0 traceMaxwell2dVzoom.eps

%% plot divergence H
everystp=20;
ttplot=tt(1:everystp:end);
vplot=vecnorm(dhPlot/Ms,2,2); vplotem=vecnorm(dhPlotem/Ms,2,2); vplotbem=vecnorm(dhPlotbem/Ms,2,2);
figure(), plot(ttplot,dx*dy*vplot(1:everystp:end),'ks-',ttplot,dx*dy*vplotbem(1:everystp:end),'k+-')
xlabel('Time','FontSize',14);
ylabel('Norm(div)','Rotation',0,'HorizontalAlignment','right','FontSize',14);
legend('SEXP','SEM','Location','NorthWest')
%  str=sprintf('$M_s=$ %d, $\\Delta t=$ %.2f, $M=$ %d',Ms,dt,M);
%  title(str,'Interpreter','latex','FontSize',14)
axis([0 T -.01 .01])
print -depsc2 -r0 normdivMaxwell2dV-R.eps

save('traceMaxwell2dVerwer.mat')


























